PROGRAM Main

   USE mod_types
   USE mod_common_var
   USE mod_interface, ONLY: r, sub_exp_mu, sub_rnumderiv, &
        sub_valueiter, sub_eulererrors, sub_simuldyneuler
   USE mod_felicity
  
   IMPLICIT NONE

   REAL (dp), DIMENSION(len_a_grid)		:: c_guess
   REAL (dp), allocatable :: yp_grid(:), pi_yp(:,:), picum_yp(:,:)
   REAL (dp), allocatable :: yt_grid(:), pi_yt(:,:), picum_yt(:,:)

   REAL(dp) 				:: max_euler_err
   REAL(dp), DIMENSION(len_a_grid,len_h_grid,len_y_grid) :: val, d_val, &
        opt_euler_err
   REAL(dp), DIMENSION(len_a_grid,len_h_grid,len_h_grid,len_y_grid):: &
        con,sav,val_choice,euler_errors,exp_mu
   INTEGER, DIMENSION(len_a_grid,len_h_grid,len_y_grid) :: index_house
   INTEGER	:: i1,i2,j1,j2,index,c(3)
   REAL :: t0,t1,t2,t3
   
   !Begins timing the main block.
   call cpu_time(t0)
   t1 = secnds(0.0)


   ! 1. Create grids
   !-------------------------------------------------------------
   !1.a Income grid
   !---------------------
   ALLOCATE (yp_grid(len_yp_grid),pi_yp(len_yp_grid,len_yp_grid),&
        picum_yp(len_yp_grid,len_yp_grid) )

   ALLOCATE (yt_grid(len_yt_grid),pi_yt(len_yt_grid,len_yt_grid),&
        picum_yt(len_yt_grid,len_yt_grid) )

   call sub_tauchen(yp_grid,pi_yp,picum_yp,rho_yp,mu0_yp,&
        sigma_yp,len_yp_grid,cover_yp)

   call sub_tauchen(yt_grid,pi_yt,picum_yt,rho_yt,mu0_yt,&
        sigma_yt,len_yt_grid,cover_yt)

   ! Income grid (vectorization of matrix yp_grid x yt_grid)
   do i1 = 1,len_yp_grid
      do j1 = 1,len_yt_grid
         index = len_yt_grid*(i1-1) + j1
         y_grid(index) = yp_grid(i1) + yt_grid(j1)
      end do
   end do

   y_grid = exp(y_grid)

   ! Transition matrix for income
   do i1 = 1, size(pi_yp,1)
      do i2 = 1, size(pi_yt,1)
         do j1 = 1, size(pi_yp,2)
            do j2 = 1, size(pi_yt,2)
               pi_y((i1-1)*size(pi_yt,1)+i2,(j1-1)*size(pi_yt,2)+j2) = &
                    pi_yp(i1,j1)*pi_yt(i2,j2)
            end do
         end do
      end do
   end do

   ! Compute the CDF of the transition matrix.
   picum_y(:,1)=pi_y(:,1)
   DO i1 = 2,len_y_grid
      picum_y(:,i1)=picum_y(:,i1-1)+pi_y(:,i1)
   ENDDO

   ! 1.b Double exponential grid for assets over [min_a, max_a]
   !--------------------------
   CALL sub_grid(2,len_a_grid,min_a,max_a,a_grid)

   ! 1.c Uniform grid for housing over [min_h, max_h]
   !-------------------------
   CALL sub_grid(0,len_h_grid,min_h,max_h,h_grid)


   ! 2. Initial guess for value function and its wealth derivative
   !--------------------------------------------------------------
   DO i1 = 1,len_y_grid
      c_guess(:) =  2.d-2*a_grid(:) + y_grid(i1)
      DO i2 = 1,len_h_grid
        h_prime = h_grid(i2)
        val(:,i2,i1) = fel(c_guess)
        CALL sub_rnumderiv(a_grid,val(:,i2,i1),d_val(:,i2,i1))
      END DO
   END DO

   ! 3. Solving for value and policy fns
   !--------------------------------------------------------------
    CALL sub_valueiter(index_house,con,sav,val_choice,val,d_val)

    ! Elapsed time for the decision problem
    t2 = secnds(t1)
    call cpu_time(t3)

    PRINT *, ' '
    PRINT *, 'Elapsed cpu time in the main program =', t3-t0, 'secs'
    PRINT *, ' '

    PRINT *, ' '
    PRINT *, 'Elapsed user time in the main program =', t2, 'secs'
    PRINT *, ' '
    
    
    ! 4. Calculating Euler errors at nodes (non-zero only if endogenous
    ! grid points are not used)
    !------------------------------------------------------------
    
    ! Solves for future marginal utility at nodes
    CALL sub_exp_mu(con,sav,index_house,val_choice,exp_mu)

    ! Solves for Euler errors at nodes
    CALL sub_eulererrors(con,exp_mu,euler_errors)    
    euler_errors = log10(abs(euler_errors))
    DO i1 = 1,len_y_grid
       DO i2 = 1,len_h_grid
          DO j1= 1,len_a_grid
             opt_euler_err(j1,i2,i1) = &
                  euler_errors(j1,index_house(j1,i2,i1),i2,i1)
          END DO
       END DO
    END DO
    
    max_euler_err = MAXVAL(opt_euler_err,mask=opt_euler_err<7.d0)
    PRINT *, "Max Euler error",max_euler_err
    c = MAXLOC(opt_euler_err,mask=opt_euler_err<7.d0)
    print *, c

    ! 5. Calculates max and average Euler error along a simulated path.
    !-------------------------------------------------------------------

    CALL sub_simuldyneuler(con,sav,val_choice,picum_y)

 END PROGRAM main
