MODULE mod_felicity
! Contain the felicity function (fel) and its first (dfel) and second
! derivative (d2fel).  
! The three of them need to be entered by the USER.  
! Each function has a scalar (_sc) and an array (_ar) version. 
! They both need to be entered.
  USE mod_types
  USE mod_common_var  

  IMPLICIT NONE

  INTERFACE fel
     MODULE PROCEDURE fel_sc, fel_ar
  END INTERFACE

  INTERFACE dfel
     MODULE PROCEDURE dfel_sc, dfel_ar
  END INTERFACE

  INTERFACE d2fel
     MODULE PROCEDURE d2fel_sc, d2fel_ar
  END INTERFACE

CONTAINS

  !******************************************************
  !               Felicity function                     !
  !******************************************************
  FUNCTION fel_sc(con)
    !Returns felicity for a consumption SCALAR 

    REAL(dp), INTENT(IN)        :: con
    REAL(dp)                    :: fel_sc

    IF (tau.NE.0.d0) THEN
      fel_sc = LOG(theta*con**tau+(1.d0-theta)*(kappa*h_prime)**tau)/tau
    ELSE
      fel_sc = theta*LOG(con)+(1.d0-theta)*LOG(kappa*h_prime)
    END IF

  END FUNCTION fel_sc


  FUNCTION fel_ar(con)
    !Returns felicity ARRAY for a consumption ARRAY

    REAL(dp), INTENT(IN), DIMENSION(:)	:: con
    REAL(dp), DIMENSION(SIZE(con))	:: fel_ar

    IF (tau.NE.0.d0) THEN
      fel_ar = LOG(theta*con**tau+(1.d0-theta)*(kappa*h_prime)**tau)/tau
    ELSE
      fel_ar = theta*LOG(con)+(1.d0-theta)*LOG(kappa*h_prime)
    END IF

  END FUNCTION fel_ar

  !******************************************************
  !            Marginal felicity function               !
  !******************************************************
  FUNCTION dfel_sc(con)
    !Returns marginal felicity for a consumption SCALAR

    REAL(dp), INTENT(IN)        :: con
    REAL(dp)                    :: dfel_sc

    dfel_sc = theta*con**(tau-1.d0)/(theta*con**tau+&
        (1.d0-theta)*(kappa*h_prime)**tau)

  END FUNCTION dfel_sc

  FUNCTION dfel_ar(con)
    !Returns marginal felicity for a consumption ARRAY

    REAL(dp), INTENT(IN), DIMENSION(:)	:: con
    REAL(dp), DIMENSION(SIZE(con))	:: dfel_ar

    dfel_ar = theta*con**(tau-1.d0)/(theta*con**tau+&
        (1.d0-theta)*(kappa*h_prime)**tau)

  END FUNCTION dfel_ar

  !******************************************************
  !      Second derivative of felicity function         !
  !******************************************************
  FUNCTION d2fel_sc(con)
    !Returns second der. of felicity for a consumption scalar

    REAL(dp), INTENT(IN)        :: con
    REAL(dp)                    :: d2fel_sc

    d2fel_sc = (-tau*((theta*con**(tau-1))/&
        (theta*con**tau+(1-theta)*(kappa*h_prime)**tau))**2 +&
        (tau-1)*theta*con**(tau-2)/&
        (theta*con**tau+(1-theta)*(kappa*h_prime)**tau)) 

  END FUNCTION d2fel_sc

  FUNCTION d2fel_ar(con)
    !Returns second der. of felicity for a consumption ARRAY

    REAL(dp), INTENT(IN), DIMENSION(:)	:: con
    REAL(dp), DIMENSION(SIZE(con))	:: d2fel_ar

    d2fel_ar = (-tau*((theta*con**(tau-1))/&
        (theta*con**tau+(1-theta)*(kappa*h_prime)**tau))**2 +&
        (tau-1)*theta*con**(tau-2)/&
        (theta*con**tau+(1-theta)*(kappa*h_prime)**tau)) 

  END FUNCTION d2fel_ar

END MODULE mod_felicity
