MODULE mod_interface

  !1.Interface to array-valued functions
  !--------------------------------------------------------
  
  INTERFACE r
     FUNCTION r_sc(a,index_h)
       USE mod_types
       USE mod_common_var
       IMPLICIT NONE
       INTEGER, INTENT(in) 	:: index_h
       REAL(dp), INTENT(in) :: a
       REAL(dp)			:: r_sc
     END FUNCTION r_sc

     FUNCTION r_ar(a,index_h)
       USE mod_types
       USE mod_common_var
       IMPLICIT NONE
       INTEGER, INTENT(in) 			:: index_h
       REAL(dp), INTENT(in), DIMENSION(:)  	:: a
       REAL (dp), DIMENSION(SIZE(a))		:: r_ar
     END FUNCTION r_ar
   END INTERFACE

!2.Interface to subroutines
!--------------------------------------------------------

   INTERFACE
      SUBROUTINE sub_eulererrors(con,exp_mu,euler_errors)
        USE mod_types
        USE mod_common_var
        IMPLICIT NONE
        REAL(dp), INTENT(in), DIMENSION(:,:,:,:)	:: con,exp_mu
        REAL(dp), INTENT(out), &
             DIMENSION(SIZE(con,1),SIZE(con,2),SIZE(con,3),SIZE(con,4)) :: &
             euler_errors
      END SUBROUTINE sub_eulererrors
   END INTERFACE

  INTERFACE
     SUBROUTINE sub_exp_mu(con,sav,index_house,val_choice,exp_mu)
       USE mod_types
       USE mod_common_var
       IMPLICIT NONE
       REAL(dp), INTENT(in), DIMENSION(:,:,:,:)	:: con, sav, val_choice
       INTEGER, INTENT(in), DIMENSION(:,:,:)		:: index_house
       REAL(dp), INTENT(out), &
           DIMENSION(SIZE(con,1),SIZE(con,2),SIZE(con,3),SIZE(con,4)) :: exp_mu
     END SUBROUTINE sub_exp_mu
  END INTERFACE

  INTERFACE
     SUBROUTINE sub_grid(n_case,len_grid,min,max,grid_out)
       USE mod_types
       USE mod_common_var  
       IMPLICIT NONE
       INTEGER, INTENT(in) 	:: n_case, len_grid
       REAL (dp), INTENT(in)	:: min, max
       REAL (dp), INTENT(out), DIMENSION(len_grid) :: grid_out
     END SUBROUTINE sub_grid
  END INTERFACE

  INTERFACE
     SUBROUTINE sub_poloffgrid(c_grid,s_grid,val_grid,a,h_out,c,s)
       USE mod_types
       USE mod_common_var
       IMPLICIT NONE
       REAL(dp), INTENT(in), DIMENSION(len_a_grid,len_h_grid):: &
           c_grid, s_grid, val_grid
       REAL(dp), INTENT(in)			:: a
       INTEGER, INTENT(out) 			:: h_out
       REAL(dp), INTENT(out)			:: c, s
     END SUBROUTINE sub_poloffgrid
  END INTERFACE

  INTERFACE
     SUBROUTINE sub_rnumderiv(X,Y,dY_dX)
       USE mod_types
       IMPLICIT NONE
       REAL(dp), DIMENSION(:),INTENT(in) :: X,Y
       REAL(dp), DIMENSION(SIZE(X)),INTENT(out) :: dY_dX
       REAL(dp), DIMENSION(SIZE(X)) :: X_plus,dX,Y_plus,dY
     END SUBROUTINE sub_rnumderiv
  END INTERFACE

  INTERFACE
     SUBROUTINE sub_simul(euler_errors,sav,val_choice,transitioncdf)
       USE mod_types
       USE mod_common_var
       IMPLICIT NONE
       REAL(dp), INTENT(in), &
           DIMENSION(len_a_grid,len_h_grid,len_h_grid,len_y_grid):: &
           euler_errors,sav,val_choice
       REAL(dp), INTENT(in), DIMENSION(len_y_grid,len_y_grid) :: transitioncdf
     END SUBROUTINE sub_simul
  END INTERFACE

  INTERFACE
     SUBROUTINE sub_simuldyneuler(con,sav,val_choice,transitioncdf)
       USE mod_types
       USE mod_common_var
       USE, INTRINSIC:: ieee_arithmetic
       IMPLICIT NONE
       REAL(dp), INTENT(in), &
           DIMENSION(len_a_grid,len_h_grid,len_h_grid,len_y_grid)::&
           con,sav,val_choice
       REAL(dp), INTENT(in), DIMENSION(len_y_grid,len_y_grid) :: transitioncdf
     END SUBROUTINE sub_simuldyneuler
  END INTERFACE
  
  INTERFACE
     SUBROUTINE sub_valueiter &
         (index_house,con,sav,val_choice,val_in,dval_in)
       USE mod_types
       USE mod_common_var  
       IMPLICIT NONE
       REAL (dp), INTENT(in), DIMENSION(len_a_grid,len_h_grid&
           &,len_y_grid):: val_in, dval_in
       REAL (dp), INTENT(out), DIMENSION(len_a_grid,len_h_grid,len_h_grid,&
           len_y_grid) 	:: con, sav, val_choice
       INTEGER, INTENT(out), DIMENSION(len_a_grid,len_h_grid,len_y_grid)::&
           index_house
     END SUBROUTINE sub_valueiter
  END INTERFACE

END MODULE mod_interface
