SUBROUTINE sub_simuldyneuler(con,sav,val_choice,transitioncdf)
  ! Calculates Euler errors along a simulated path. Averages over
  ! all points where Euler equation holds.
  ! Adapted from sub_simul by Francisco Barillas and 
  ! Jesus Fernandez-Villaverde - undated				 
		
  USE mod_types
  USE mod_common_var
  USE, intrinsic:: ieee_arithmetic
  USE mod_interface, ONLY: r,sub_poloffgrid
  USE mod_fcn
  USE mod_felicity
  USE mod_newton

  IMPLICIT NONE

  REAL(dp), INTENT(in), &
       DIMENSION(len_a_grid,len_h_grid,len_h_grid,len_y_grid)::&
       con,sav,val_choice
  REAL(dp), INTENT(in), DIMENSION(len_y_grid,len_y_grid) :: transitioncdf

  REAL(dp), DIMENSION(len_y_grid) :: c_2, fut_mu, s_2
  REAL(dp), DIMENSION(Tsimul) 	  :: eerrorsim
  REAL(dp) :: c_1, c_out, exp_mu, max_err, meaneerror,s_1, pippo, &
      randomnumber, wealth
  INTEGER :: h_1,i,index_h,index_y,j(1),time,nonzero_time,nonzero_t2
  INTEGER, DIMENSION(len_y_grid) :: h_2
  
  !Variables used by sub_newton subroutine
  INTEGER, PARAMETER    :: N=1
  REAL(dp)              :: FNORM, F(N), X(N), XGUESS(N)

  !Specifies the initial state
  wealth = 0.d0 
  index_h = 1
  index_y = INT(len_y_grid/2)+1

  nonzero_time = 0
  nonzero_t2 = 0
  eerrorsim = 1.d-15
  DO time=1,Tsimul
    CALL sub_poloffgrid(con(:,:,index_h,index_y),&
        sav(:,:,index_h,index_y),val_choice(:,:,index_h,index_y),wealth,&
        h_1,c_1,s_1)
    IF (s_1.ge.a_grid(2)) THEN 
      DO i =1, len_y_grid
        CALL sub_poloffgrid(con(:,:,h_1,i),&
            sav(:,:,h_1,i),val_choice(:,:,h_1,i),&
            s_1,h_2(i),c_2(i),s_2(i))
        h_prime = h_grid(h_2(i))
        fut_mu(i) = dfel(c_2(i))
      END DO
      exp_mu = DOT_PRODUCT(pi_y(index_y,:),fut_mu(:))
      ! global variables used by the fn euler
      h_prime = h_grid(h_1)
      b_global = bita*(1+r(wealth,h_1))*exp_mu
      xguess = c_1
      CALL sub_newton(sub_fcn1,sub_d2fcn,xguess,x,0.d0,1.d-8,1.d-8)
      c_out = x(1)
      eerrorsim(time) = 1.d0 - c_out/c_1
      nonzero_time = nonzero_time + 1
      IF (time>10000) THEN
         nonzero_t2 = nonzero_t2+1
      END IF
      
      IF (ieee_is_nan(eerrorsim(time))) THEN
        PRINT *, s_1, h_1, index_h, index_y
        ! Pause
        WRITE(*,'("  1 ")') !BP
        READ(*,*) 
      END IF

    END IF

    wealth = s_1
    index_h = h_1
    CALL RANDOM_NUMBER(randomnumber)
    j = MINLOC(transitioncdf(index_y,:), &
        mask=transitioncdf(index_y,:)>randomnumber)
    index_y=j(1)
  ENDDO
  
  eerrorsim = LOG10(ABS(eerrorsim))
  max_err = MAXVAL(eerrorsim)
  meaneerror=SUM(PACK(eerrorsim(:),eerrorsim>-15.d0))/REAL(nonzero_time,dp)
	
  PRINT *, ' '
  PRINT *, 'The maximum euler equation error on the simulated path is'
  PRINT *, max_err


  PRINT *, ' '
  PRINT *, 'The average euler equation error on the simulated path is'
  PRINT *, real(nonzero_time,dp), meaneerror

  PRINT *, ' '
  PRINT *, 'The average euler equation error on the last T - 10000 periods is'
  PRINT *, REAL(nonzero_t2,dp), SUM(PACK(eerrorsim(10001:),&
       eerrorsim(10001:)>-15.d0))/REAL(nonzero_t2,dp)

end subroutine sub_simuldyneuler
