SUBROUTINE sub_eulererrors(con,exp_mu,euler_errors)
  ! Calculates Euler errors given 
  ! 1) present consumption (con) and housing (h)
  ! 2) expected future marginal felicity
  ! Euler errors are conditional on present housing choice (same as con)
  ! i.e. they are a function of the state AND housing choice

  USE mod_types
  USE mod_common_var
  USE mod_interface, ONLY: r
  USE mod_fcn
  USE mod_newton

  IMPLICIT NONE
  
  REAL(dp), INTENT(in), DIMENSION(:,:,:,:)	:: con,exp_mu
  REAL(dp), INTENT(out), &
       DIMENSION(SIZE(con,1),SIZE(con,2),SIZE(con,3),SIZE(con,4)) :: &
       euler_errors

  REAL(dp)					:: a_prime, c_out, fut_con
  REAL(dp), DIMENSION(len_h_grid)		:: val
  REAL(dp), DIMENSION(len_y_grid)		:: fut_mu
  INTEGER		:: i, index_a, index_h, index_h_prime, index_y,info
  !Variables used by sub_newton subroutine
  INTEGER, PARAMETER    :: N=1
  REAL(dp)              :: FNORM, F(N), X(N), XGUESS(N)

   euler_errors = 1.d7
	
	
  DO index_y = 1, len_y_grid
     DO index_h = 1, len_h_grid
        DO index_h_prime = 1, len_h_grid
           DO index_a = 1, len_a_grid
              xguess  = con(index_a,index_h_prime,index_h,index_y)
              ! Compute Euler errors only when not borrowing constrained
              ! This is the case iff exp_mu(.)>0
              IF (exp_mu(index_a,index_h_prime,index_h,index_y )>0.d0) THEN
                 ! global variables used by the fn euler
                 h_prime = h_grid(index_h_prime)
                 b_global = bita*(1+r(a_grid(index_a),index_h))*&
                      exp_mu(index_a,index_h_prime,index_h,index_y)
                 CALL sub_newton(sub_fcn1,sub_d2fcn,xguess,x,0.d0)
                 c_out = x(1)
                 euler_errors(index_a,index_h_prime,index_h,index_y) = &
                      1- c_out/con(index_a,index_h_prime,index_h,index_y)
              END IF
           
           END DO
        END DO
     END DO
  END DO
  
END SUBROUTINE sub_eulererrors
